#ifndef OUTPUTFILE_H
#define OUTPUTFILE_H

#include <tbb/concurrent_queue.h>

#include <atomic>
#include <cstdint>
#include <cstdio>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <string>
#include <string_view>
#include <thread>
#include <utility>
#include <vector>

#include "cmssw/FRDFileHeader_v2.h"
#include "log.h"

namespace fs = std::filesystem;

class OutputFile {
 public:
  OutputFile(std::string_view filepath, std::string_view filename, std::string_view run_dir,
             FRDFileHeader_v2 header, uint32_t num_lumisection, uint32_t file_index)
      : filepath_(filepath),
        fileName_(filename),
        rundir_(run_dir),
        fileHeader_(header),
        lumisection_(num_lumisection),
        index_(file_index) {
    try {
      file_.open(filepath_, std::ios_base::out | std::ios_base::binary);
    } catch (std::exception& e) {
      LOG(FATAL) << "Exception opening file" << filepath_ << "', exception: " << e.what();
    }
  }

  OutputFile(const OutputFile&) = delete;
  OutputFile& operator=(const OutputFile&) = delete;

  OutputFile(OutputFile&& other) noexcept
      : filepath_(std::move(other.filepath_)),
        fileName_(std::move(other.fileName_)),
        rundir_(std::move(other.rundir_)),
        fileHeader_(std::move(other.fileHeader_)),
        lumisection_(other.lumisection_),
        index_(other.index_) {
    file_ = std::move(other.file_);
  }

  OutputFile& operator=(OutputFile&& other) noexcept {
    if (this != &other) {
      filepath_ = std::move(other.filepath_);
      fileName_ = std::move(other.fileName_);
      rundir_ = std::move(other.rundir_);
      fileHeader_ = std::move(other.fileHeader_);
      lumisection_ = other.lumisection_;
      index_ = other.index_;
      file_ = std::move(other.file_);
    }
    return *this;
  }

  std::string getFileName() { return fileName_; }

  FRDFileHeader_v2& getFileHeader() { return fileHeader_; }
  std::string getRunDir() { return rundir_; }

  bool writeFileHeader() {
    file_.seekp(0, std::ios::beg);
    file_.write(reinterpret_cast<const char*>(&fileHeader_), sizeof(fileHeader_));
    return !file_.fail();
  }

  void Seek(size_t pos) {
    file_.seekp(pos, std::ios::beg);  // Move write position
  }

  std::string GetFullPath() const { return filepath_; }

  void Rename(const std::string& new_path) {
    fs::path from = GetFullPath();
    fs::path to = new_path;

    LOG(TRACE) << "Renaming from '" << from << "' to '" << to << "' now.";

    try {
      fs::rename(from, to);
    } catch (fs::filesystem_error& err) {
      LOG(FATAL) << "Error renaming file: " << err.what();
    }

    filepath_ = new_path;
  }

  size_t Write(char* buf, size_t size) {
    file_.write(buf, size);
    return size;
  }

  bool IsOpen() { return file_.is_open(); }

  bool Close() {
    if (IsOpen()) file_.close();
    return !IsOpen();
  }

  void setFileHeader(FRDFileHeader_v2 fileHeader) { fileHeader_ = fileHeader; }

  uint32_t Lumisection() { return lumisection_; }
  uint32_t Index() { return index_; }

 private:
  std::string filepath_;
  std::string fileName_;
  std::string rundir_;
  FRDFileHeader_v2 fileHeader_;
  uint32_t lumisection_;
  uint32_t index_;

  std::fstream file_;
  //  FILE* filePtr_;
};

#endif
