#include "OutputFileHandler.h"

#include <cassert>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <memory>

#include "OutputFile.h"
#include "log.h"
#include "tools.h"

const std::string OutputFileHandler::working_dir_ = "in_progress";
const std::string OutputFileHandler::journal_file_ = "index.journal";

tbb::concurrent_bounded_queue<std::shared_ptr<OutputFile>> OutputFileHandler::files_to_close_ =
    tbb::concurrent_bounded_queue<std::shared_ptr<OutputFile>>();
std::atomic<bool> OutputFileHandler::file_handler_running_{true};

void OutputFileHandler::enqueue_current_file_for_close_and_move_maybe() {
  if (outputFile_ && outputFile_->IsOpen()) {
    LOG(TRACE) << "queueing file: " << outputFile_->getFileName() << " for closing, queue size now "
               << files_to_close_.size();
    if (this->OutputFileHandler::getCMSSWHeaders()) {
      outputFile_->setFileHeader(createFileHeader(outputFile_->Lumisection()));
      outputFile_->writeFileHeader();
    }
    ls_file_size_ += file_size_;
    ls_NOrbits_ += NOrbits_;
    if (outputFile_->getFileHeader().getFileSize() > 32) {
      files_to_close_.push(outputFile_);
      LOG(TRACE) << "queued " << outputFile_->getFileName() << " queue size now "
                 << files_to_close_.size();
    } else {
      outputFile_->Close();
    }
    // resetting for next file
    NOrbits_ = 0;
    if (this->OutputFileHandler::getCMSSWHeaders()) {
      file_size_ = sizeof(FRDFileHeader_v2);
    } else {
      file_size_ = 0;
    }
    outputFile_ = nullptr;
  }
}

bool OutputFileHandler::MaybeCommitFile(uint32_t run, uint32_t index) {
  bool is_new_run = (current_run_number_ != static_cast<int>(run));
  bool is_new_index = (current_index_ != static_cast<int>(index));

  if (current_run_number_ >= 0 && is_new_run) {
    // Write by default even without cmssw headers present
    write_EoR_file();
  }

  bool do_commit_file = is_new_run || is_new_index;

  if (do_commit_file) {
    enqueue_current_file_for_close_and_move_maybe();
  }

  if (is_new_run) {
    LOG(TRACE) << "Previous run: " << current_run_number_ << " | New run: " << run;
    LOG(TRACE) << "Previous index: " << current_index_ << " | New index: " << index;

    run_NOrbits_ = 0;
    current_run_number_ = static_cast<int>(run);
    current_index_ = static_cast<int>(index);
    run_dir_ = base_path_ + "/" + GetRunFormatted(current_run_number_);
  } else if (is_new_index) {
    LOG(TRACE) << "Previous index: " << current_index_ << " | New index: " << index;
    current_index_ = static_cast<int>(index);
  }

  return do_commit_file;
}

std::shared_ptr<OutputFile>& OutputFileHandler::GetFile(uint32_t run, uint32_t index) {
  if (MaybeCommitFile(run, index)) {
    open_new_file();
  }
  return outputFile_;
}

FRDFileHeader_v2 OutputFileHandler::createFileHeader(uint32_t ls) {  // file header for CMSSW input
  uint16_t headerSize = sizeof(FRDFileHeader_v2);
  uint16_t dataType = static_cast<uint16_t>(sourceID_);
  uint32_t eventCount = NOrbits_;
  uint32_t runNumber = current_run_number_;
  uint64_t fileSize = file_size_;
  FRDFileHeader_v2 fileHeader(headerSize, dataType, eventCount, runNumber, ls, fileSize);
  return fileHeader;
}

void OutputFileHandler::create_output_directory_maybe(std::string& output_directory) {
  struct stat sb;
  LOG(TRACE) << "checking if working directory " << output_directory << " exists ";
  /* check if path exists and is a directory */
  if (stat(output_directory.c_str(), &sb) == 0) {
    if (S_ISDIR(sb.st_mode)) {  // output directory already exists
      return;
    }
    std::string err = "ERROR The output directory path '" + output_directory +
                      "' exists, but the path is not a directory!";
    LOG(ERROR) << err;
    throw std::runtime_error(err);
  }

  if (!tools::filesystem::create_directories(output_directory)) {
    std::string err =
        tools::strerror("ERROR when creating the output directory '" + output_directory + "'");
    LOG(ERROR) << err;
    throw std::runtime_error(err);
  }
  LOG(TRACE) << "Created output directory: " << output_directory << "'.";
}

void OutputFileHandler::open_new_file() {
  std::string working_files_base_path = run_dir_ + "/" + working_dir_;
  // Create a new file
  uint32_t ls = 1 + static_cast<uint32_t>(current_index_ / (max_index_per_ls_ + 1));
  create_output_directory_maybe(working_files_base_path);
  std::string filename = format_filename(current_run_number_, current_index_, ls);
  std::string full_filename = working_files_base_path + "/" + filename;
  LOG(TRACE) << "opening file with index " << current_index_ << ", in lumisection " << ls;

  outputFile_ =
      std::make_shared<OutputFile>(full_filename, filename, run_dir_, createFileHeader(ls), ls,
                                   current_index_ % (max_index_per_ls_ + 1));
  // reserve space for CMSSW header if required
  if (this->OutputFileHandler::getCMSSWHeaders()) {
    outputFile_->Seek(sizeof(FRDFileHeader_v2));
  }
}

// Create a properly formatted file name
std::string OutputFileHandler::format_filename(uint32_t run_number, uint32_t index, uint32_t ls) {
  std::ostringstream ofilename;

  if (this->OutputFileHandler::getCMSSWHeaders()) {
    ofilename << GetRunFormatted(run_number) << "_" << GetLumiFormatted(ls) << "_"
              << GetFileIndexFormatted(index % (max_index_per_ls_ + 1)) << filename_suffix_;

  } else {
    ofilename << filename_prefix_ << "_" << std::setfill('0') << std::setw(6) << run_number << "_"
              << std::setfill('0') << std::setw(6) << index << filename_suffix_;
  }
  return ofilename.str();
}

void OutputFileHandler::close_and_rename::operator()() const {
  std::shared_ptr<OutputFile> outputFile;
  while (file_handler_running_ || (files_to_close_.size() > 0)) {
    LOG(TRACE) << "try pop now. queue size now " << files_to_close_.size();
    try {
      files_to_close_.pop(outputFile);
    } catch (tbb::user_abort& e) {
      LOG(ERROR) << "Exception caught when trying to pop files to close queue.";
      break;
    }
    LOG(TRACE) << "popping file: " << outputFile->getFileName() << " for closing, queue size now "
               << files_to_close_.size();

    if (outputFile->Close()) {
      LOG(ERROR) << tools::strerror("File close failed");
    }

    std::string permanent_path(outputFile->getRunDir() + "/" + outputFile->getFileName());
    outputFile->Rename(permanent_path);

    if (outputFile->Index() == outputfilehandler_->max_index_per_ls_ &&
        outputfilehandler_->getCMSSWHeaders()) {  // end of lumisection
      outputfilehandler_->write_EoLS_file(outputFile->Lumisection());
    }
  }
  LOG(TRACE) << "exiting close_and_rename. file_handler_running_ now " << file_handler_running_
             << " queue size now " << files_to_close_.size();
}

void OutputFileHandler::write_EoLS_file(uint32_t ls) {
  std::stringstream EoLS_filename_stringstream;
  EoLS_filename_stringstream << GetRunFormatted(current_run_number_) << "_" << GetLumiFormatted(ls)
                             << "_EoLS.jsn";
  std::string EoLS_filename = EoLS_filename_stringstream.str();
  std::string full_filename = run_dir_ + "/" + EoLS_filename;
  std::stringstream EoLS_file_string_stream;
  std::fstream EoLS_file;
  LOG(TRACE) << "writing EoLS file " << EoLS_filename;
  EoLS_file.open(full_filename.c_str(), std::ios_base::out);

  // TODO: To me it looks like there are more fields than intended in this file...
  EoLS_file_string_stream << "{\n  \"data\":[\"" << ls_NOrbits_ << "\",\""  // NEvents
                          << 1 + max_index_per_ls_ << "\",\""               // NFiles
                          << constants::N_orbits_per_lumisection << "\","   // Total Events
                          << "\"0\",\""                                     // NLost Events
                          << ls_file_size_ << "\"],\n  \"definition\":\""   // NBytes
                          << "/fff/ramdisk/" << GetRunFormatted(current_run_number_)
                          << "/jsd/EoLS.jsd\",\n  \"source\":\"l1scout\"\n}";

  ls_file_size_ = 0;
  run_NOrbits_ += ls_NOrbits_;
  ls_NOrbits_ = 0;
  EoLS_file << EoLS_file_string_stream.str();
  EoLS_file.close();
}

void OutputFileHandler::write_EoR_file() {
  assert(current_index_ > 0 && nOrbitsPerFile_ > 0);
  int ls = int(1) + (current_index_ * nOrbitsPerFile_) / constants::N_orbits_per_lumisection;
  std::stringstream EoR_filename;
  EoR_filename << run_dir_ << "/" << GetRunFormatted(current_run_number_) << "_"
               << GetLumiFormatted(0) << "_EoR.jsn";
  LOG(TRACE) << "Writing EoR file " << EoR_filename.str();
  std::fstream EoR_file;
  EoR_file.open(EoR_filename.str().c_str(), std::ios_base::out);
  std::stringstream EoR_file_string_stream;
  EoR_file_string_stream << "{\n  \"data\":[\"" << run_NOrbits_  // NEvents
                         << "\",\""
                         << (ls * static_cast<int>(max_index_per_ls_ + 1)) +
                                (nOrbitsPerFile_ * current_index_)  // NFiles
                         << "\",\"" << ls << "\",\""                // NLumis
                         << ls << "\"],\n  \"definition\":\""       // LastLumi
                         << "/fff/ramdisk/" << GetRunFormatted(current_run_number_)
                         << "/jsd/EoR.jsd\",\n  \"source\":\"l1scout\"\n}";
  EoR_file << EoR_file_string_stream.str();
  EoR_file.close();
}
